/*++

Copyright (c) Microsoft Corporation.  All rights reserved.

    THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
    KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
    PURPOSE.

Module Name:

    Toaster.c

Abstract:

    This is a simple form of function driver for toaster device. The driver
    doesn't handle any PnP and Power events because the framework provides
    default behavior for those events. This driver has enough support to
    allow an user application (toast/notify.exe) to open the device
    interface registered by the driver and send read, write or ioctl requests.

Environment:

    Kernel mode

--*/

#include "toaster.h"
#include "..\common\IchIoControl.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, ToasterEvtDeviceAdd)
#pragma alloc_text (PAGE, ToasterEvtIoDeviceControl)
#endif


NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
/*++

Routine Description:
    DriverEntry initializes the driver and is the first routine called by the
    system after the driver is loaded. DriverEntry configures and creates a WDF driver
    object.
    .
Parameters Description:

    DriverObject - represents the instance of the function driver that is loaded
    into memory. DriverObject is allocated by the system before the
    driver is loaded, and it is released by the system after the system unloads
    the function driver from memory.

    RegistryPath - represents the driver specific path in the Registry.
    The function driver can use the path to store driver related data between
    reboots. The path does not store hardware instance specific data.

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise.

--*/
{
    NTSTATUS            status = STATUS_SUCCESS;
    WDF_DRIVER_CONFIG   config;
	PDEVICE_OBJECT      deviceObject = NULL;

    KdPrint(("Toaster Function Driver Sample - Driver Framework Edition.\n"));
    KdPrint(("Built %s %s\n", __DATE__, __TIME__));

    //
    // Initiialize driver config to control the attributes that
    // are global to the driver. Note that framework by default
    // provides a driver unload routine. If DriverEntry creates any resources
    // that require clean-up in driver unload,
    // you can manually override the default by supplying a pointer to the EvtDriverUnload
    // callback in the config structure. In general xxx_CONFIG_INIT macros are provided to
    // initialize most commonly used members.
    //

    WDF_DRIVER_CONFIG_INIT(
        &config,
        ToasterEvtDeviceAdd
        );


    //
    // Create a framework driver object to represent our driver.
    //
    status = WdfDriverCreate(
        DriverObject,
        RegistryPath,
        WDF_NO_OBJECT_ATTRIBUTES, // Driver Attributes
        &config,          // Driver Config Info
        WDF_NO_HANDLE
        );

    if (!NT_SUCCESS(status)) {
        KdPrint( ("WdfDriverCreate failed with status 0x%x\n", status));
    }

    return status;
}


NTSTATUS
ToasterEvtDeviceAdd(
    IN WDFDRIVER       Driver,
    IN PWDFDEVICE_INIT DeviceInit
    )
/*++
Routine Description:

    ToasterEvtDeviceAdd is called by the framework in response to AddDevice
    call from the PnP manager. We create and initialize a WDF device object to
    represent a new instance of toaster device.

Arguments:

    Driver - Handle to a framework driver object created in DriverEntry

    DeviceInit - Pointer to a framework-allocated WDFDEVICE_INIT structure.

Return Value:

    NTSTATUS

--*/
{
    NTSTATUS               status = STATUS_SUCCESS;
    PFDO_DATA              fdoData;
    WDF_IO_QUEUE_CONFIG    queueConfig;
    WDF_OBJECT_ATTRIBUTES  fdoAttributes;
    WDFDEVICE              hDevice;
    WDFQUEUE               queue;

    UNREFERENCED_PARAMETER(Driver);

    PAGED_CODE();

    KdPrint(("ToasterEvtDeviceAdd called\n"));

    //
    // Initialize attributes and a context area for the device object.
    //
    //
    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&fdoAttributes, FDO_DATA);

    //
    // Create a framework device object.This call will in turn create
    // a WDM device object, attach to the lower stack, and set the
    // appropriate flags and attributes.
    //
    status = WdfDeviceCreate(&DeviceInit, &fdoAttributes, &hDevice);
    if (!NT_SUCCESS(status)) {
        KdPrint( ("WdfDeviceCreate failed with status code 0x%x\n", status));
        return status;
    }

    //
    // Tell the Framework that this device will need an interface
    //
    status = WdfDeviceCreateDeviceInterface(
                 hDevice,
                 (LPGUID) &GUID_DEVINTERFACE_TOASTER,
                 NULL // ReferenceString
             );

    if (!NT_SUCCESS (status)) {
        KdPrint( ("WdfDeviceCreateDeviceInterface failed 0x%x\n", status));
        return status;
    }

    //
    // Register I/O callbacks to tell the framework that you are interested
    // in handling IRP_MJ_READ, IRP_MJ_WRITE, and IRP_MJ_DEVICE_CONTROL requests.
    // If a specific callback function is not specified for one ofthese,
    // the request will be dispatched to the EvtIoDefault handler, if any.
    // If there is no EvtIoDefault handler, the request will be failed with
    // STATUS_INVALID_DEVICE_REQUEST.
    // WdfIoQueueDispatchParallel means that we are capable of handling
    // all the I/O requests simultaneously and we are responsible for protecting
    // data that could be accessed by these callbacks simultaneously.
    // A default queue gets all the requests that are not
    // configured for forwarding using WdfDeviceConfigureRequestDispatching.
    //
    WDF_IO_QUEUE_CONFIG_INIT_DEFAULT_QUEUE(&queueConfig,  WdfIoQueueDispatchParallel);

    queueConfig.EvtIoDeviceControl = ToasterEvtIoDeviceControl;

    status = WdfIoQueueCreate(
        hDevice,
        &queueConfig,
        WDF_NO_OBJECT_ATTRIBUTES,
        &queue
        );

    if (!NT_SUCCESS (status)) {

        KdPrint( ("WdfIoQueueCreate failed 0x%x\n", status));
        return status;
    }

    return status;
}

/*****************************************************************************/
NTSTATUS
ReadWriteConfigSpace(
    __in PDEVICE_OBJECT DeviceObject,
    __in ULONG          ReadOrWrite, // 0 for read 1 for write
    __in PVOID          Buffer,
    __in ULONG          Offset,
    __in ULONG          Length
    )
/*++
Routine Description:

    Worker function to read & write to PCI config space at
    PASSIVE_LEVEL

Arguments:

    DeviceObject - Pointer to the PDO
    ReadOrWrite  - Whichspace - 0 for read 1 for write
    Buffer       - Buffer to be written to or read from
    Offset       - Offset in to the PCI config space
    Length       - Length of the buffer

Return Value:

     NT Status code

--*/
{
    KEVENT event;
    NTSTATUS status;
    PIRP irp;
    IO_STATUS_BLOCK ioStatusBlock;
    PIO_STACK_LOCATION irpStack;
    PDEVICE_OBJECT targetObject;

    PAGED_CODE();

    KeInitializeEvent( &event, NotificationEvent, FALSE );

    targetObject = IoGetAttachedDeviceReference( DeviceObject );

    irp = IoBuildSynchronousFsdRequest( IRP_MJ_PNP,
                                        targetObject,
                                        NULL,
                                        0,
                                        NULL,
                                        &event,
                                        &ioStatusBlock );

    if (irp == NULL) {
        status = STATUS_INSUFFICIENT_RESOURCES;
        goto End;
    }

    irpStack = IoGetNextIrpStackLocation( irp );
    if (ReadOrWrite == 0) {
        irpStack->MinorFunction = IRP_MN_READ_CONFIG;
    }else {
        irpStack->MinorFunction = IRP_MN_WRITE_CONFIG;
    }

    irpStack->Parameters.ReadWriteConfig.WhichSpace = PCI_WHICHSPACE_CONFIG;
    irpStack->Parameters.ReadWriteConfig.Buffer = Buffer;
    irpStack->Parameters.ReadWriteConfig.Offset = Offset;
    irpStack->Parameters.ReadWriteConfig.Length = Length;

    //
    // Initialize the status to error in case the bus driver does not
    // set it correctly.
    irp->IoStatus.Status = STATUS_NOT_SUPPORTED ;

    status = IoCallDriver( targetObject, irp );

    if (status == STATUS_PENDING) {

        KeWaitForSingleObject( &event, Executive, KernelMode, FALSE, NULL );
        status = ioStatusBlock.Status;

    }

End:
    // Done with reference
    ObDereferenceObject( targetObject );
    return status;
}

/****************************************************************************************/
NTSTATUS ReadGpioBase (PDEVICE_OBJECT DeviceObject, USHORT *BaseAddress)
{
	//USHORT Vendor;
	//USHORT Device;
	ULONG  GpioBase;
	ULONG  GpioControl;

	NTSTATUS               status = STATUS_SUCCESS;

	/*status = ReadWriteConfigSpace (DeviceObject, 0x0, &Vendor, 0,sizeof(Vendor));
	if (!NT_SUCCESS (status))	
	{
		KdPrint (("ReadGpioBase: ReadWriteConfigSpace failed. status=%d\n",status));
		return -1;
	}
	KdPrint (("Vendor=0x%04x\n",Vendor));

	status = ReadWriteConfigSpace (DeviceObject, 0x0, &Device, 0x2,sizeof(Device));
	if (!NT_SUCCESS (status))	
	{
		KdPrint (("ReadGpioBase: ReadWriteConfigSpace failed. status=%d\n",status));
		return -1;
	}
	KdPrint (("Device=0x%04x\n",Device));*/

	//read GPIOBASE from config. space
	status = ReadWriteConfigSpace (DeviceObject, 0x0, &GpioBase, GPIO_BASE_ADDRESS_OFFSET,sizeof(GpioBase));
	if (!NT_SUCCESS (status))	
	{
		KdPrint (("ReadGpioBase: ReadWriteConfigSpace failed. status=%d\n",status));
		return -1;
	}

	*BaseAddress = GpioBase & 0x0000FF80;
	//KdPrint (("GpioBase=0x%x\n",*BaseAddress));

	//read GPIO control from config. space
	status = ReadWriteConfigSpace (DeviceObject, 0x0, &GpioControl, GPIO_CONTROL_OFFSET,sizeof(GpioControl));
	if (!NT_SUCCESS (status))	
	{
		KdPrint (("ReadGpioBase: ReadWriteConfigSpace failed. status=%d\n",status));
		return -1;
	}

	//KdPrint (("GpioControl=0x%x\n",GpioControl));

	return status;
}

/****************************************************************************************/
VOID
ToasterEvtIoDeviceControl(
    IN WDFQUEUE     Queue,
    IN WDFREQUEST   Request,
    IN size_t       OutputBufferLength,
    IN size_t       InputBufferLength,
    IN ULONG        IoControlCode
    )
/*++
Routine Description:

    This event is called when the framework receives IRP_MJ_DEVICE_CONTROL
    requests from the system.

Arguments:

    Queue - Handle to the framework queue object that is associated
            with the I/O request.
    Request - Handle to a framework request object.

    OutputBufferLength - length of the request's output buffer,
                        if an output buffer is available.
    InputBufferLength - length of the request's input buffer,
                        if an input buffer is available.

    IoControlCode - the driver-defined or system-defined I/O control code
                    (IOCTL) that is associated with the request.

Return Value:

   VOID

--*/
{
    NTSTATUS  status= STATUS_SUCCESS;
	ULONG               bytesRead =0;
	PFDO_DATA           fdoData;
	USHORT				IoBaseAddress;
	WDFDEVICE           device;
	PDEVICE_OBJECT		DeviceObject;
	ICH_WRITE_PORT_LONG_REQUEST	*pWriteRequest;
	ICH_READ_PORT_LONG_REQUEST	*pReadRequest;
	ICH_READ_PORT_LONG_REPLY *pReadReply;
	size_t              bytesReturned = 0;

	void *pInBuffer;
	void *pOutBuffer;
	unsigned long Length;

    UNREFERENCED_PARAMETER(Queue);
    UNREFERENCED_PARAMETER(OutputBufferLength);
    UNREFERENCED_PARAMETER(InputBufferLength);

    //KdPrint(("--> ToasterEvtIoDeviceControl\n"));

    PAGED_CODE();

	//
    // initialize variables
    //
    device = WdfIoQueueGetDevice(Queue);

	DeviceObject = WdfDeviceWdmGetDeviceObject (device);

    //
    // Use WdfRequestRetrieveInputBuffer and WdfRequestRetrieveOutputBuffer
    // to get the request buffers.
    //

	ReadGpioBase (DeviceObject, &IoBaseAddress);
	//KdPrint (("IoControlCode=%d IoBaseAddress=%x\n",IoControlCode,IoBaseAddress));

	switch (IoControlCode) 
	{
	case ICH_WRITE_PORT_LONG_CODE:
		status = WdfRequestRetrieveInputBuffer (Request,sizeof(ICH_WRITE_PORT_LONG_REQUEST),
												&pInBuffer, &Length);
		if (!NT_SUCCESS (status))
			KdPrint (("ToasterEvtIoDeviceControl: WdfRequestRetrieveInputBuffer failed. status=0x%x\n",status));

		if (Length!=sizeof(ICH_WRITE_PORT_LONG_REQUEST))
			KdPrint (("ToasterEvtIoDeviceControl: WdfRequestRetrieveInputBuffer - Length=%d\n",Length)); 

		pWriteRequest = (ICH_WRITE_PORT_LONG_REQUEST *)pInBuffer;

		WRITE_PORT_ULONG (IoBaseAddress + pWriteRequest->Offset,  pWriteRequest->Data);
		
		//KdPrint(("Write 0x%x to 0x%x\n",pWriteRequest->Data,pWriteRequest->Offset));

		bytesReturned = 0;
		WdfRequestCompleteWithInformation (Request,STATUS_SUCCESS,bytesReturned);
		break;

	case ICH_READ_PORT_LONG_CODE:
		status = WdfRequestRetrieveInputBuffer (Request,sizeof(ICH_READ_PORT_LONG_REQUEST),
												&pInBuffer, &Length);
		if (!NT_SUCCESS (status))
			KdPrint (("ToasterEvtIoDeviceControl: WdfRequestRetrieveInputBuffer failed. status=0x%x\n",status));

		if (Length!=sizeof(ICH_READ_PORT_LONG_REQUEST))
			KdPrint (("ToasterEvtIoDeviceControl: WdfRequestRetrieveInputBuffer - Length=%d\n",Length)); 

		pReadRequest = (ICH_READ_PORT_LONG_REQUEST *)pInBuffer;

		status = WdfRequestRetrieveOutputBuffer (Request,sizeof(ICH_READ_PORT_LONG_REPLY),
										&pOutBuffer, &Length);
		if (!NT_SUCCESS (status))
			KdPrint (("ToasterEvtIoDeviceControl: WdfRequestRetrieveOutputBuffer failed. status=0x%x\n",status));

		if (Length!=sizeof(ICH_READ_PORT_LONG_REPLY))
			KdPrint (("ToasterEvtIoDeviceControl: WdfRequestRetrieveInputBuffer - Length=%d\n",Length)); 

		pReadReply = (ICH_READ_PORT_LONG_REPLY *)pOutBuffer;
		
		pReadReply->Data = READ_PORT_ULONG (IoBaseAddress+pReadRequest->Offset);

		//KdPrint(("Read 0x%x from 0x%x\n",pReadReply->Data,pReadRequest->Offset));

		bytesReturned = sizeof(ICH_READ_PORT_LONG_REPLY);
		WdfRequestCompleteWithInformation (Request,STATUS_SUCCESS,bytesReturned);
		break;

    default:
        status = STATUS_INVALID_DEVICE_REQUEST;
    }

	//KdPrint(("<-- ToasterEvtIoDeviceControl\n"));
}


